#   Copyright 2023 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import inspect
from contextlib import AsyncExitStack, asynccontextmanager, contextmanager
from functools import partial
from itertools import chain
from typing import Any, Callable, Generic, Iterable, Optional, Type, TypeVar, Union

from eventyst.adapters.kafka import EventBroker
from eventyst.core.commands import Command
from eventyst.core.events import BaseMessage, Event
from eventyst.core.message_bus import MessageBus
from eventyst.utilities.function_inspection import lenient_issubclass

T = TypeVar("T")


class Dependency(Generic[T]):
    parameter_name_message_bus: Optional[str]

    def __init__(
        self,
        dependency: Callable[..., T],
        *,
        cache: bool = True,
        emitted_types: Iterable[Type[BaseMessage]] | None = None,
    ):
        self.dependency = dependency
        self.cache = cache
        self.emitted_types = set(emitted_types or [])

    def __call__(self, *args, **kwargs) -> Any:
        if self.dependency is None:
            raise TypeError("Dependency not set")
        return self.dependency(*args, **kwargs)


class DependencyTracker:
    def __init__(
        self,
        *,
        name: Optional[str] = None,
        call: Callable[..., Any],
        cache: bool = True,
        parameter_name_message_bus: Optional[str] = None,
        parameter_name_context: Optional[str] = None,
        parameter_name_event_broker: Optional[str] = None,
        payload_name: Optional[str] = None,
        payload_type: Optional[Type[Event] | Type[Command]] = None,
        dependencies: Optional[dict[str, 'DependencyTracker']] = None,
        emitted_types: Iterable[Type[BaseMessage]] | None = None,
    ) -> None:
        self.call = call
        self.cache = cache
        self.name = name
        self.parameter_name_message_bus = parameter_name_message_bus
        self.parameter_name_context = parameter_name_context
        self.parameter_name_event_broker = parameter_name_event_broker
        self.payload_name = payload_name
        self.payload_type = payload_type
        self.dependencies = dependencies or {}
        self.emitted_types = set(emitted_types or [])


def dependency(
    function: Callable[..., Any] = None, *, cache: bool = True, emitted_types: Iterable[Type[BaseMessage]] | None = None
) -> Any:
    emitted_types = chain(emitted_types or [], getattr(function, "__emitted_types__", []))
    if function is None:
        return partial(dependency, cache=cache, emitted_types=emitted_types)
    return Dependency(function, cache=cache, emitted_types=emitted_types)


def get_handler_tracker(
    function: Callable[..., T], emitted_types: Iterable[Type[BaseMessage]] | None = None
) -> DependencyTracker:
    return build_dependency_tracker(dependency(function), emitted_types=emitted_types)


def build_dependency_tracker(
    dependency: Dependency, name: Optional[str] = None, emitted_types: Iterable[Type[BaseMessage]] | None = None
) -> DependencyTracker:
    if not isinstance(dependency, Dependency):
        raise TypeError("Input must be an instance of Dependency")

    tracker = DependencyTracker(
        call=dependency.dependency,
        cache=dependency.cache,
        name=name,
        emitted_types=emitted_types,
    )

    signature = inspect.signature(dependency.dependency)
    for name, param in signature.parameters.items():
        annotation, sub_dependency = analyze_signature_parameter(name, param)
        if sub_dependency is not None:
            evaluated_dependency = build_dependency_tracker(
                sub_dependency, name=name, emitted_types=sub_dependency.emitted_types
            )
            tracker.emitted_types.update(evaluated_dependency.emitted_types)
            tracker.dependencies[name] = evaluated_dependency

        if annotation != inspect.Parameter.empty:
            set_default_param_names(name, annotation, tracker)

    # Check the return type and if its an event, set the emitted type
    if signature.return_annotation != inspect.Signature.empty:
        annotation = signature.return_annotation
        if lenient_issubclass(annotation, Event):
            tracker.emitted_types.add(annotation)
    # Iterate through dependencies and ensure that there is only a single payload type
    payload_types = set()
    for sub_tracker in tracker.dependencies.values():
        if sub_tracker.payload_type is not None:
            payload_types.add(sub_tracker.payload_type)

    if tracker.payload_type is not None:
        payload_types.add(tracker.payload_type)

    if len(payload_types) > 1:
        raise ValueError(f"Multiple Payload types defined for {tracker.call}\n")

    if tracker.payload_type is None and len(payload_types) > 0:
        tracker.payload_type = payload_types.pop()

    return tracker


async def solve_dependency_tracker(
    *,
    tracker: DependencyTracker,
    payload: Union[Command, Event],
    stack: AsyncExitStack,
    context: Optional[dict[str, Any]] = None,
    message_bus: Optional[MessageBus] = None,
    overrides: Optional[dict[Type, Any]] = None,
    cache: Optional[dict[Callable[..., Any], Any]] = None,
) -> tuple[dict[str, Any], dict[Callable[..., Any], Any]]:
    """Iterates through the dependency tracker and returns a the inputs for the underlying callable."""
    if not isinstance(tracker, DependencyTracker):
        raise TypeError("Input must be an instance of DependencyTracker")

    cache = cache or {}
    overrides = overrides or {}
    inputs: dict[str, Any] = {}
    # iterate through the dependency tracker and solve all sub-dependencies recursively
    for name, sub_tracker in tracker.dependencies.items():
        sub_inputs, sub_cache = await solve_dependency_tracker(
            tracker=sub_tracker,
            payload=payload,
            cache=cache,
            message_bus=message_bus,
            stack=stack,
            overrides=overrides,
        )

        # update the cache
        cache.update(sub_cache)
        # Inject the default values
        if sub_tracker.parameter_name_event_broker is not None:
            assert sub_tracker.parameter_name_event_broker is not None, "Event Broker not set"
            sub_inputs[sub_tracker.parameter_name_event_broker] = overrides[EventBroker]

        if sub_tracker.parameter_name_message_bus is not None:
            sub_inputs[sub_tracker.parameter_name_message_bus] = message_bus

        if sub_tracker.parameter_name_context is not None:
            sub_inputs[sub_tracker.parameter_name_context] = context

        if sub_tracker.payload_name is not None:
            sub_inputs[sub_tracker.payload_name] = payload

        call = sub_tracker.call
        call = overrides.get(call, call)
        # Check cache for dependency
        if tracker.call in cache:
            sub_output = cache[call]
        elif inspect.isasyncgenfunction(call):
            cm = asynccontextmanager(call)(**sub_inputs)
            sub_output = await stack.enter_async_context(cm)
        elif inspect.isgeneratorfunction(call):
            cm = contextmanager(call)(**sub_inputs)
            sub_output = stack.enter_context(cm)
        elif inspect.iscoroutinefunction(call):
            sub_output = await call(**sub_inputs)
        else:
            sub_output = call(**sub_inputs)

        # If the dependency has a name it is an input to the parent dependency
        if sub_tracker.name is not None:
            inputs[name] = sub_output

        # If the dependency is not cached yet, add it to the cache
        if sub_tracker.cache and sub_tracker.call not in cache:
            cache[sub_tracker.call] = sub_output

    if tracker.parameter_name_event_broker is not None:
        assert tracker.parameter_name_event_broker is not None, "Event Broker not set"
        inputs[tracker.parameter_name_event_broker] = overrides[EventBroker]

    # Inject the default values for the input dependency
    if tracker.parameter_name_message_bus is not None:
        inputs[tracker.parameter_name_message_bus] = message_bus

    if tracker.parameter_name_context is not None:
        inputs[tracker.parameter_name_context] = context

    if tracker.payload_name is not None:
        inputs[tracker.payload_name] = payload

    return inputs, cache


def set_default_param_names(param_name: str, annotation: Any, tracker: DependencyTracker) -> DependencyTracker:
    """Search through the signature of the dependency's callable and set the param names that map to default values."""

    if lenient_issubclass(annotation, MessageBus):
        if tracker.parameter_name_message_bus:
            raise ValueError(
                f"Cannot have multiple message buses as a dependency for {tracker.call}\n"
                f"parameters {tracker.parameter_name_message_bus} and {param_name} both define a subclass of MessageBus"
            )
        tracker.parameter_name_message_bus = param_name
    elif lenient_issubclass(annotation, (Event, Command)):
        if tracker.payload_name is not None:
            raise ValueError(
                f"Cannot have multiple events or commands as a dependency for {tracker.call}\n"
                f"parameters {tracker.payload_name} and {param_name} both define a subclass of Event or Command"
            )
        tracker.payload_name = param_name
        tracker.payload_type = annotation
    elif lenient_issubclass(annotation, EventBroker):
        if tracker.parameter_name_event_broker is not None:
            raise ValueError(
                f"Cannot have multiple events or commands as a dependency for {tracker.call}\n"
                f"parameters {tracker.payload_name} and {param_name} both define a subclass of Event or Command"
            )
        tracker.parameter_name_event_broker = param_name
    return tracker


_RESERVED_INJECTION_TYPES = (MessageBus, Event, Command, EventBroker)


def analyze_signature_parameter(parameter_name: str, parameter: inspect.Parameter) -> tuple[Any, Optional[Dependency]]:
    # Set default outputs
    annotation = Any
    dependency = None
    # Check if the default value is a dependency
    if isinstance(parameter.default, Dependency):
        dependency = parameter.default
    # Extract the annotation
    if parameter.annotation != inspect.Parameter.empty:
        annotation = parameter.annotation

    if dependency is not None and lenient_issubclass(annotation, _RESERVED_INJECTION_TYPES):
        raise ValueError(
            f"Cannot use {dependency} as a dependency for {parameter_name} because it is a reserved injection type"
        )
    return annotation, dependency



#   Copyright 2023 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import asyncio
import inspect
import logging
from contextlib import AsyncExitStack
from functools import partial
from typing import TYPE_CHECKING, Callable, Dict, Iterable, List, ParamSpec, Type, TypeVar

from eventyst.core import commands, events
from eventyst.core.dependency import DependencyTracker, get_handler_tracker, solve_dependency_tracker

if TYPE_CHECKING:
    from eventyst.adapters.kafka import EventBroker
    from eventyst.service_layer.message_bus import MessageBus

P = ParamSpec('P')
R = TypeVar('R')

logger = logging.getLogger(__name__)


class HandlerRegistry:
    command_handlers: Dict[Type[commands.Command], DependencyTracker]
    event_handlers: Dict[Type[events.Event], List[DependencyTracker]]

    def __init__(
        self,
        event_handlers: Dict[Type[events.Event], List[DependencyTracker]] | None = None,
        command_handlers: Dict[Type[commands.Command], DependencyTracker] | None = None,
        broker_events: set[Type[events.Event]] | None = None,
    ):
        self.event_handlers = event_handlers or {}
        self.command_handlers = command_handlers or {}
        self.broker_events = broker_events or set()
        self.entrypoints = set()
        self.overrides = {}

    def register(
        self,
        handler: Callable[P, R],
        entrypoint: bool = True,
        emitted_types: Iterable[Type[events.BaseMessage]] | None = None,
    ) -> Callable[P, R]:
        tracker = get_handler_tracker(handler, emitted_types)
        if tracker.payload_type is None:
            raise ValueError(f"Handler {handler} does not have a Command or Event as a parameter")
        elif issubclass(tracker.payload_type, events.Event):
            self.register_events({tracker.payload_type: [tracker]})
        elif issubclass(tracker.payload_type, commands.Command):
            self.register_commands({tracker.payload_type: tracker})
        else:
            raise ValueError(f"Handler {handler} has unsupported payload type {tracker.payload_type}")

        if entrypoint:
            self.entrypoints.add(tracker.payload_type)
        return handler

    def register_commands(self, handlers: Dict[Type[commands.Command], DependencyTracker]):
        # check for duplicate command handlers
        duplicate_handlers = set(handlers.keys()) & set(self.command_handlers.keys())
        if duplicate_handlers:
            raise ValueError(f"Duplicate command handlers for commands {duplicate_handlers}")
        self.command_handlers.update(handlers)

    def register_events(self, handlers: Dict[Type[events.Event], List[DependencyTracker]]):
        for event_type, event_handlers in handlers.items():
            if event_type not in self.event_handlers:
                self.event_handlers[event_type] = []
            self.event_handlers[event_type].extend(event_handlers)

    def get_event_handlers(self, event: events.Event) -> List[DependencyTracker]:
        return self.event_handlers.get(type(event), [])

    def get_command_handler(self, command: commands.Command) -> DependencyTracker:
        if type(command) not in self.command_handlers:
            raise ValueError(f"No handler registered for commands of type {type(command)}")
        return self.command_handlers[type(command)]

    async def handle_event(
        self,
        event: events.Event,
        bus: 'MessageBus',
        broker: 'EventBroker | None' = None,
    ):
        loop = asyncio.get_running_loop()
        tasks = []
        for handler in self.get_event_handlers(event):
            tasks.append(loop.create_task(self.call_handler(handler, event, bus)))

        if event.__class__ in self.broker_events and broker is not None:
            logger.debug(f"Producing event {event} to broker")
            tasks.append(loop.create_task(broker.produce(event)))
        return await asyncio.gather(*tasks)

    async def call_handler(
        self,
        handler: DependencyTracker,
        event: events.Event,
        bus: 'MessageBus',
    ):
        loop = asyncio.get_running_loop()
        async with AsyncExitStack() as stack:
            try:
                inputs, _ = await solve_dependency_tracker(
                    tracker=handler,
                    payload=event,
                    message_bus=bus,
                    stack=stack,
                    overrides=self.overrides,
                )
                if inspect.iscoroutinefunction(handler.call):
                    output = await handler.call(**inputs)
                else:
                    output = await loop.run_in_executor(None, partial(handler.call, **inputs))
                if output is not None and isinstance(output, events.Event):
                    await bus.enqueue(output)
            except Exception:
                logger.exception(f"Exception handling event {event}")

    async def handle_command(
        self,
        command: commands.Command,
        bus: 'MessageBus',
    ):
        loop = asyncio.get_running_loop()
        handler = self.get_command_handler(command)
        async with AsyncExitStack() as stack:
            inputs, _ = await solve_dependency_tracker(
                tracker=handler,
                payload=command,
                message_bus=bus,
                stack=stack,
                overrides=self.overrides,
            )
            if inspect.iscoroutinefunction(handler.call):
                output = await handler.call(**inputs)
            else:
                output = await loop.run_in_executor(None, partial(handler.call, **inputs))
            if output is not None and isinstance(output, events.Event):
                await bus.enqueue(output)

    def override_dependency(
        self,
        dependency: Type,
        override,
    ):
        self.overrides[dependency] = override



#   Copyright 2023 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
Settings needed for each Eventyst mircoservice to communicate with one another.

This includes the ports for the API and for Kafka as well as all of the
information needed to access the database.
"""

import os
from textwrap import dedent

from pydantic import BaseSettings, PostgresDsn, SecretStr, parse_obj_as

from eventyst._enums import LogLevel

# PREFIX for all environment variables
_PACKAGE_PREFIX = "EVENTYST_"


class PostgresqlDsn(PostgresDsn):
    """Restricts allowed schema for PostgresDsn Pydantic Model to postgresql."""

    allowed_schemes = {"postgresql+psycopg"}


class Settings(BaseSettings):
    """
    Eventyst Settings Management Object that carries the global configuration for the Eventyst application.

    Inherits from the Pydantic BaseSettings object (https://pydantic-docs.helpmanual.io/usage/settings/) to set the
    configuration from environment variables, dotenv files and defaults.
    """

    SERVICE_NAME: str = "eventyst"
    APPLICATION_PATH: str = 'model::app'
    # Kafka Settings
    KAFKA_BOOTSTRAP_SERVERS: str = "localhost:9092"

    # Local Settings
    LOG_LEVEL: LogLevel = LogLevel.INFO
    # Database Credentials
    POSTGRES_DSN: PostgresqlDsn = parse_obj_as(PostgresqlDsn, "postgresql+psycopg://postgres@localhost/eventyst")
    POSTGRES_PASSWORD: SecretStr = SecretStr("")
    POSTGRES_SCHEMA: str = "public"
    ENGINE_ECHO: bool = False
    APP_STR: str = "model.main:app"
    # AWS Settings
    AWS_SCHEMA_REGISTRY_NAME: str = "eventyst"
    AWS_SCHEMA_REGISTRY_REGION: str = "us-east-2"

    class Config:
        """Pydantic Configuration."""

        env_file = os.environ.get("EVENTYST_ENV_FILE", ".env")
        env_prefix = _PACKAGE_PREFIX

    def display(self, show_defaults: bool = False, show_passwords: bool = False):
        """Display a valid dotenv file version of the settings for human readability and logging."""
        params = []
        for key, val in self.dict().items():
            if val is not None:
                str_val = f"{val.get_secret_value()}" if show_passwords and "PASSWORD" in key else val
                if show_defaults or key in self.__fields_set__:
                    params.append(f"{_PACKAGE_PREFIX}{key} = {str_val}")
                else:
                    params.append(f"# {_PACKAGE_PREFIX}{key} = {str_val}")

        params_str = "\n".join(params)
        output = f"""######################\n# EVENTYST Settings\n######################\n{params_str}"""
        return dedent(output)

    def __str__(self) -> str:
        """Use the human readable display method for printing."""
        return self.display()


# Instantiate the global settings for use throughout Eventyst
settings = Settings()



#   Copyright 2023 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import asyncio


def get_running_loop() -> asyncio.AbstractEventLoop | None:
    """
    Returns the running event loop. If there is no running event loop, None is returned.
    """
    try:
        return asyncio.get_running_loop()
    except RuntimeError:
        return None



#   Copyright 2023 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import logging

from rich.console import Console
from rich.logging import RichHandler

from eventyst.configuration import settings

_ROOT_LOGGER_NAME = "eventyst"

console = Console()


def setup_logger():
    logger = logging.getLogger(_ROOT_LOGGER_NAME)
    logger.setLevel(settings.LOG_LEVEL.get_log_level())
    handler = RichHandler(console=console)
    handler.setLevel(settings.LOG_LEVEL.get_log_level())
    logger.addHandler(handler)
    return logger


def get_logger():
    return logging.getLogger(_ROOT_LOGGER_NAME)


def get_child_logger(name: str):
    # check if name begins with root logger name
    if name.startswith(_ROOT_LOGGER_NAME):
        # get root
        root = get_logger()
        return root.getChild(name[len(_ROOT_LOGGER_NAME) + 1 :])

    return logging.getLogger(name)


root = setup_logger()



#   Copyright 2023 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Abstract and concrete schema classes."""

import abc
import json
from dataclasses import dataclass
from io import BytesIO
from typing import Any, Literal, Optional, Union
from uuid import UUID

import fastavro
import fastavro.types

SchemaVersionStatus = Literal['AVAILABLE', 'DELETING', 'FAILURE', 'PENDING']
SchemaFormat = Literal['AVRO', 'JSON']


class BaseSchema(abc.ABC):
    @property
    @abc.abstractmethod
    def data_format(self) -> SchemaFormat:
        """The data format of this schema."""

    @property
    @abc.abstractmethod
    def fully_qualified_name(self) -> str:
        """The fully-qualified name of this schema."""

    @abc.abstractmethod
    def read(self, bytes_: bytes, reader_schema: Optional["BaseSchema"] = None) -> Any:
        """Read bytes into a record."""

    @abc.abstractmethod
    def write(self, data) -> bytes:
        """Write a record into bytes."""

    @abc.abstractmethod
    def validate(self, data) -> None:
        """Raise a ValidationException if the data is invalid."""

    @abc.abstractmethod
    def json(self) -> str:
        """Return the JSON representation of this schema."""

    def __hash__(self) -> int:
        return hash(str(self))

    def __eq__(self, other) -> bool:
        return (
            isinstance(other, BaseSchema)
            and self.data_format == other.data_format
            and self.fully_qualified_name == other.fully_qualified_name
        )


class AvroSchema(BaseSchema):
    """An Avro schema."""

    _definition: dict
    _parsed: fastavro.types.Schema

    def __init__(self, definition: Union[str, dict]) -> None:
        """Create a new AvroSchema."""
        definition = json.loads(definition) if isinstance(definition, str) else definition
        self._definition = definition
        self._parsed = fastavro.parse_schema(definition)

    def __str__(self) -> str:
        return self.json()

    @property
    def data_format(self) -> SchemaFormat:
        return 'AVRO'

    @property
    def fully_qualified_name(self) -> str:
        # https://github.com/fastavro/fastavro/issues/415
        return self._parsed.get('name', self._parsed['type'])

    def read(self, bytes_: bytes, reader_schema: Optional[fastavro.types.Schema] = None) -> Any:
        b = BytesIO(bytes_)
        value = fastavro.schemaless_reader(
            b,
            self._parsed,
            reader_schema=reader_schema._parsed,
        )
        b.close()
        return value

    def write(self, data) -> bytes:
        b = BytesIO()
        fastavro.schemaless_writer(b, self._parsed, data)
        value = b.getvalue()
        b.close()
        return value

    def validate(self, data) -> None:
        fastavro.validate(data, self._parsed)

    def json(self) -> str:
        return json.dumps(self._parsed)

    def __getitem__(self, key: str):
        return self._definition[key]


@dataclass
class SchemaVersion:
    schema_name: str
    version_id: UUID
    definition: str
    data_format: SchemaFormat
    status: SchemaVersionStatus
    version_number: Optional[int] = None

    def __hash__(self):
        return hash(self.definition)



#   Copyright 2023 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import asyncio
import threading
from typing import Any, Callable, Iterable, ParamSpec, Type, TypeVar

import boto3
from fastapi import Depends, FastAPI

from eventyst.adapters.aws.schema_registry import GlueSchemaRegistryClient
from eventyst.adapters.kafka import KafkaEventBroker
from eventyst.configuration import settings
from eventyst.core.commands import Command
from eventyst.core.events import BaseMessage, Event
from eventyst.core.handler_registry import HandlerRegistry
from eventyst.core.message_bus import MessageBus
from eventyst.serialization.serializers import DataAndSchema, Deserializer, Serializer
from eventyst.utilities.log import get_child_logger

logger = get_child_logger(__name__)

In = ParamSpec("In")
Out = TypeVar("Out")


class Eventyst:
    def __init__(self, name: str = "Eventyst", version: str = "0.1.0", registry: HandlerRegistry | None = None) -> None:
        self.name = name
        self.version = version
        self.registry = registry or HandlerRegistry()
        self.broker = None
        self.logger = get_child_logger(f"eventyst.{self.name}")

    def build_api(self):
        # iterate through registry and register routes
        app = FastAPI(title=self.name, version=self.version)

        @app.get("/")
        async def root():
            return {"message": "Hello World"}

        @app.get("/health")
        async def health():
            return {"message": "OK"}

        bus = MessageBus(self.registry)

        def message_bus():
            return bus

        async def submit_event(
            event: Event,
            bus: MessageBus = Depends(message_bus),
        ):
            await bus.handle(event)
            return event

        # override submit_event's event parameter to be an AnalysisRan
        for handler in self.registry.command_handlers.values():
            if handler.payload_type is None:
                continue
            submit_event.__annotations__["event"] = handler.payload_type
            app.post(
                f"/command/{handler.call.__name__.replace('_','-')}",
                summary=f"Submit a {handler.payload_type.__name__} Command",
                tags=["Command"],
            )(submit_event)
        return app, bus

    def register_with_broker(self, event: Type[Event]):
        # check if event is already registered
        if event in self.registry.broker_events:
            raise Exception(f"{event} is already registered with the broker")
        # check if event is an Event
        if not issubclass(event, Event):
            raise Exception(f"You can only register Events with the broker. {event} is not an Event")
        self.registry.broker_events.add(event)
        return event

    def start_api(self):
        app, bus = self.build_api()
        thread = threading.Thread(target=bus.start, daemon=True)
        thread.start()
        return app

    def produce(self, *events: BaseMessage, topic: str | None = None, raw_event: dict | None = None):
        logger.debug(f"Producing {len(events)} to {topic}")
        asyncio.run(self._single_producer(events, topic))

    def _initalize_broker(self):
        if self.broker is not None:
            return
        session = boto3.Session(region_name=settings.AWS_SCHEMA_REGISTRY_REGION)
        glue_client = session.client('glue')
        registry_client = GlueSchemaRegistryClient(
            glue_client=glue_client,
            registry_name=settings.AWS_SCHEMA_REGISTRY_NAME,
        )
        serializer = Serializer(
            client=registry_client,
        )
        deserializer = Deserializer(client=registry_client, reader_schema=BaseMessage.avro_schema())
        self.broker = KafkaEventBroker(
            settings.KAFKA_BOOTSTRAP_SERVERS,
            serializer=serializer,
            deserializer=deserializer,
        )

    async def _single_producer(self, events: BaseMessage, topic: str | None = None):
        self._initalize_broker()
        assert self.broker is not None, "Broker should be initialized"
        async with self.broker:
            for event in events:
                await self.broker.produce(event, topic)

    async def start_consumer(
        self,
        payload_type: Type[Command] | Type[Event],
        auto_offset_reset: str = "latest",
        group_id: str = "eventyst",
    ):
        session = boto3.Session()
        glue_client = session.client('glue')
        registry_client = GlueSchemaRegistryClient(
            glue_client=glue_client,
            registry_name=settings.AWS_SCHEMA_REGISTRY_NAME,
        )

        logger.info(f"Starting consumer for {payload_type.topic()}")
        serializer = Serializer(
            client=registry_client,
        )
        deserializer = Deserializer(client=registry_client, reader_schema=payload_type.avro_schema())
        broker = KafkaEventBroker(
            settings.KAFKA_BOOTSTRAP_SERVERS,
            serializer=serializer,
            deserializer=deserializer,
        )
        bus = MessageBus(self.registry, event_broker=broker)
        asyncio.create_task(bus._entrypoint())
        async with broker:
            await broker.producer.start()
            consumer = await broker.get_consumer(payload_type, auto_offset_reset, group_id)
            async for message in consumer:
                try:
                    value: DataAndSchema = message.value
                    event = payload_type.parse_obj(value.data["payload"])
                    await bus.handle(event)
                except Exception as e:
                    logger.error(f"An error occurred when handling message: {e}")

    async def start_multiple_consumers(
        self,
        *payload_type_map: Type[Command] | Type[Event],
        group_id: str = "eventyst",
        auto_offset_reset: str = "latest",
    ):
        tasks = []
        for payload_type in payload_type_map:
            tasks.append(asyncio.create_task(self.start_consumer(payload_type, auto_offset_reset, group_id)))
        await asyncio.gather(*tasks)

    def register(
        self,
        handler: Callable[..., Any] | None = None,
        *,
        entrypoint: bool = False,
        emitted_types: Iterable[Type[BaseMessage]] | None = None,
    ):
        if handler:
            self.registry.register(handler, entrypoint=entrypoint, emitted_types=emitted_types)
            return handler
        else:

            def wrapper(handler_: Callable[..., Any]):
                self.registry.register(handler_, entrypoint=entrypoint, emitted_types=emitted_types)
                return handler_

            return wrapper

    def start(self, offset_reset: str = "latest", group_id: str = "eventyst"):
        if len(self.registry.entrypoints) == 0:
            raise Exception("No entrypoints registered")
        logger.info("Starting Eventyst")
        logger.info(f"Consuming from {len(self.registry.entrypoints)} topics")
        asyncio.run(
            self.start_multiple_consumers(*self.registry.entrypoints, group_id=group_id, auto_offset_reset=offset_reset)
        )



#   Copyright 2023 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import abc
from typing import AsyncIterable, Type

from aiokafka import AIOKafkaClient, AIOKafkaConsumer, AIOKafkaProducer
from aiokafka.errors import KafkaConnectionError, KafkaError

from eventyst._exceptions import BrokerConnectionError
from eventyst.core.events import BaseMessage, Event
from eventyst.serialization.registry import SchemaRegistryClient
from eventyst.serialization.serializers import Deserializer, Serializer
from eventyst.utilities.log import get_child_logger

logger = get_child_logger(__name__)


class EventBroker(abc.ABC):
    @abc.abstractmethod
    async def produce(self, event: Event, topic: str | None = None):
        pass

    @abc.abstractmethod
    async def consume(self, topic: str):
        pass

    @abc.abstractmethod
    async def get_consumer(self, payload_type: Type[BaseMessage], topic: str):
        pass


class Consumer(abc.ABC, AsyncIterable):
    @abc.abstractmethod
    async def start(self):
        pass

    @abc.abstractmethod
    async def stop(self):
        pass


class KafkaEventBroker(EventBroker):
    def __init__(
        self,
        bootstrap_servers: str,
        registry_client: SchemaRegistryClient | None = None,
        serializer: Serializer | None = None,
        deserializer: Deserializer = None,
    ):
        self.bootstrap_servers = bootstrap_servers
        self.registry_client = registry_client
        self.deserializer = deserializer
        self.serializer = serializer
        self._consumers = []

    async def test_connection(self):
        client = AIOKafkaClient(bootstrap_servers=self.bootstrap_servers)
        try:
            await client.bootstrap()
        except KafkaConnectionError:
            raise BrokerConnectionError("Could not connect to Kafka broker, please check your settings")

    async def start_producer(self):
        # check if producer is already started
        if self.producer._sender is None:
            try:
                await self.producer.start()
            except Exception as e:
                logger.debug(f"An error occurred when starting the producer: {e}")

    async def get_consumer(
        self,
        payload_type: Type[BaseMessage],
        auto_offset_reset: str = 'earliest',
        group_id: str = 'latest',
        topic: str | None = None,
    ) -> Consumer:
        topic = topic or payload_type.topic()
        # check if consumer is already started
        consumer = AIOKafkaConsumer(
            topic,
            bootstrap_servers=self.bootstrap_servers,
            auto_offset_reset=auto_offset_reset,
            group_id=group_id,
            value_deserializer=self.deserializer,
        )
        consumer.subscribe([topic])
        await consumer.start()
        return consumer

    async def produce(self, event: Event, topic: str | None = None):
        # Override topic if provided
        topic = topic or event.topic()
        await self.start_producer()
        if self.serializer is not None:
            self.producer._value_serializer = self.serializer.get_serializer(topic)
        try:
            try:
                logger.debug(f"Producing event: {event}")
                await self.producer.send(topic, (event.envelope(), event.avro_schema()))
            except KafkaError as e:
                logger.debug(f"An error occurred when sending event: {e}")
        except Exception as e:
            logger.error(f"An error occurred when sending event: {e}", exc_info=e)

    async def consume(self, topic: str, auto_offset_reset: str = "latest", group_id: str = None):
        self.consumer = AIOKafkaConsumer(
            bootstrap_servers=self.bootstrap_servers,
            value_deserializer=self.deserializer,
            auto_offset_reset=auto_offset_reset,
            group_id=group_id,
        )
        try:
            await self.consumer.start()
            try:
                self.consumer.subscribe([topic])
                async for msg in self.consumer:
                    logger.debug(f"Consumed message: {msg.value}")
                    yield msg
            except KafkaError as e:
                logger.debug(f"An error occurred when consuming message: {e}")
            finally:
                await self.consumer.stop()
        except Exception as e:
            logger.exception("An error occurred when starting the consumer", exc_info=e)

    async def __aenter__(self):
        self.producer = AIOKafkaProducer(bootstrap_servers=self.bootstrap_servers)
        await self.producer.start()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        logger.debug("Flushing and stopping producer")
        await self.producer.stop()
        logger.debug("Stopping consumers")
        for consumer in self._consumers:
            await consumer.stop()



#   Copyright 2023 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Functions to encode and decode values with schema information.

In order to benefit from using a schema registry with Kafka, key and
value data must be encoded with additional bits of information, which
effectively renders the raw bytes human unreadable.

The encoded data consists of the following components:

    byte | value
    ------------
    0    | magic byte that signifies whether the data was written with
         | a compatible client
    1    | which algorithm was used to compress the data bytes
    2-17 | uuid that indicates the writer schema
    18+  | actual data bytes, possibly compressed according to the
         | compression byte

This encoding is based on the Java Glue Schema Registry client
(https://github.com/awslabs/aws-glue-schema-registry) in an effort to
maintain full compatibility.
"""

from __future__ import annotations

import zlib
from io import BytesIO
from uuid import UUID

VERSION_BYTE = b'\x03'
"""Expected value of the magic version byte.

If the leading byte of the encoded data has a different value,
that signifies one of the following:

    1. The data was encoded by a different version of the encoder
    2. The data was encoded by a different library (e.g. the Java library)
       that is no longer compatible with this library
    3. The data was encoded for another schema registry
       (e.g. Confluent Schema Registry)
    4. The data was written by a schema-less producer
"""

COMPRESSION_ENABLED_BYTE = b'\x05'
"""Compression byte when using ZLIB compression."""

COMPRESSION_DISABLED_BYTE = b'\x00'
"""Compression byte when compression is disabled."""

SCHEMA_VERSION_ID_SIZE = 16
"""Number of bytes reserved for the schema version uuid."""


class CodecException(Exception):
    """Raised when encoding or decoding fails."""


class UnknownEncodingException(CodecException):
    """Raised when decoding data with an unknown encoding."""


def encode(data: bytes, schema_version_id: UUID, compression=None) -> bytes:
    """Encode data and schema information into bytes.

    Arguments:
        data (bytes): the payload itself.
        schema_version_id (UUID): version id of the schema used to
            serialize the data.
        compression (Any): whether to compress the payload data.
            Any truthy value can be passed to enable compression.

            Currently only ZLIB compression is supported. In future
            versions this parameter may take specific values to
            differentiate between different compression algorithms.

    Returns:
        bytes
    """
    b = BytesIO()
    if compression:
        compression_byte = COMPRESSION_ENABLED_BYTE
        data = zlib.compress(data)
    else:
        compression_byte = COMPRESSION_DISABLED_BYTE
    b.write(VERSION_BYTE)
    b.write(compression_byte)
    b.write(schema_version_id.bytes)
    b.write(data)
    value = b.getvalue()
    b.close()
    return value


def decode(bytes_: bytes) -> tuple[bytes, UUID]:
    """Decode bytes into data and schema information.

    Arguments:
        bytes_ (bytes): encoded bytes.

    Returns:
        tuple[bytes, UUID]: a two-item tuple consisting of the decoded
            and decompressed data, and the schema version id

    Raises:
        UnknownEncodingException: if the leading byte of the encoded
            data is not recognized, implying the data was encoded with
            an incompatible client or for a different schema registry
        CodecException: if any other error occurs while decoding
    """
    b = BytesIO(bytes_)
    version = b.read(1)
    if version != VERSION_BYTE:
        raise UnknownEncodingException(r"leading byte {version!r} not recognized")
    compression = b.read(1)
    schema_version = UUID(bytes=b.read(SCHEMA_VERSION_ID_SIZE))
    data = b.read()
    if compression == COMPRESSION_ENABLED_BYTE:
        data = zlib.decompress(data)
    elif compression != COMPRESSION_DISABLED_BYTE:
        raise CodecException(f'compression byte {compression!r} not recognized')
    return data, schema_version



#   Copyright 2023 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Strategies for choosing schema names in the registry."""

from typing import Protocol

from eventyst.schema import BaseSchema


class SchemaNamingStrategy(Protocol):
    """Controls how schema names are chosen for a value.

    Arguments:
        topic: the name of the topic the value is being written into.
        value: the _unserialized_ value being written.
        is_key: whether the value is a Kafka key or value.

    Returns:
        str: the schema name to use
    """

    def __call__(self, topic: str, is_key: bool, schema: BaseSchema) -> str:
        ...


def topic_name_strategy(topic: str, is_key: bool, schema: BaseSchema) -> str:
    """The default naming strategy.

    Message keys are `<topic>-key` and message values are
    `<topic>-value`.

    This is a sensible strategy for topics whose records follow a uniform
    schema, but does not allow mixing different schemas on the same topic.
    """
    return f"{topic}-{'key' if is_key else 'value'}"


def record_name_strategy(topic: str, is_key: bool, schema: BaseSchema) -> str:
    """Uses the fully-qualified record name as the schema name.

    Allows a topic to contain records with multiple incompatible schemas.
    However, this requires that the fully-qualified record names uniquely
    and consistently identify a schema across the entire registry.
    """
    return schema.fully_qualified_name


def topic_record_name_strategy(topic: str, is_key: bool, schema: BaseSchema) -> str:
    """Combines the topic and record name to form the schema name.

    Allows a topic to contain records with multiple incompatible schemas.
    Additionally allows different topics to use the same record name for
    incompatible schemas.
    """
    return f'{topic}-{schema.fully_qualified_name}'



#   Copyright 2023 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import webbrowser
from pathlib import Path
from typing import Optional

import typer

from eventyst import __version__
from eventyst._exceptions import BrokerConnectionError
from eventyst.cli import styles
from eventyst.cli.utilities import validate_app_str
from eventyst.configuration import settings
from eventyst.utilities.plotting import draw_graph_pyvis, generate_graph

app = typer.Typer(no_args_is_help=True)


app_str_option = typer.Option(None, '--app', envvar='EVENYTST_APP_STR', help="The application to draw")


@app.command(name="version")
def version(short: bool = typer.Option(False, "-s", help="Print the raw version with no ascii art.")):
    """Display gcld version info"""
    if short:
        styles.console.print(__version__)
    else:
        styles.console.print(styles.LOGO_STYLE)


@app.command(name="config")
def get_config(
    show_password: bool = False,
    show_defaults: bool = False,
    out_pth: Optional[Path] = typer.Option(
        None,
        "--out",
        "-o",
        help="Location to write parametrized config",
    ),
):
    """
    Prints out the configuration of gcld
    """
    # If out_pth provided write the current config to the path provided and return
    if out_pth:
        with open(out_pth, "w") as f:
            f.write(settings.display(True, True))

    styles.theme_typer_print(settings.display(show_defaults, show_password))


@app.command(name="run")
def run(
    app_str: str = app_str_option,
    offset_reset: str = typer.Option('latest', help="Offset reset for consumer"),
):
    """
    Run a Eventyst application
    """
    app_str = app_str or settings.APP_STR
    app = validate_app_str(app_str)
    try:
        app.start(offset_reset)
    except BrokerConnectionError as e:
        app.logger.error(f"Could not connect to broker: {e}")
        raise typer.Exit(1)
    except KeyboardInterrupt:
        app.logger.info("Shutting down...")
        raise typer.Exit(0)
    except Exception as e:
        app.logger.exception(f"An error occurred when starting the application: {e}", exc_info=e)
        raise typer.Exit(1)


@app.command(name="draw")
def draw(
    app_str: str = app_str_option,
    out_pth: Optional[Path] = typer.Option('example.html', "--out", "-o", help="Location to write html file"),
    buttons: list[str] = typer.Option(None, "--buttons", "-b", help="Buttons to show in the graph"),
    all_buttons: bool = typer.Option(False, "--all-buttons", "-A", help="Show all buttons in the graph"),
    auto_open: bool = typer.Option(False, "--open", "-O", help="Open the graph in a browser"),
):
    """
    Draw the application
    """
    app_str = app_str or settings.APP_STR
    app = validate_app_str(app_str)

    G = generate_graph(app.registry)
    net = draw_graph_pyvis(G)
    # open in browser
    if all_buttons:
        net.show_buttons()
    elif buttons:
        net.show_buttons(filter_=buttons)

    net.show(str(out_pth))
    if auto_open:
        webbrowser.open(str(out_pth))


@app.command(name="start-api")
def start_api(
    app_str: str = app_str_option,
):
    """
    Start the api for the application
    """
    app_str = app_str or settings.APP_STR
    app = validate_app_str(app_str)
    app.start_api()



#   Copyright 2023 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import asyncio
from typing import TYPE_CHECKING, Optional, Union

from eventyst.core import commands, events
from eventyst.utilities.log import get_child_logger

if TYPE_CHECKING:
    from eventyst.adapters.kafka import EventBroker
    from eventyst.core.handler_registry import HandlerRegistry

logger = get_child_logger(__name__)


class MessageBus:
    def __init__(
        self,
        handler_registry: 'HandlerRegistry',
        event_broker: 'Optional[EventBroker]' = None,
    ):
        self.handler_registry = handler_registry

        # setup queue and loop
        self.queue: asyncio.Queue[Union[commands.Command, events.Event]] = asyncio.Queue()
        self.submissions = set()

        self.event_broker = event_broker

        # Events
        self._ready_event = asyncio.Event()
        self._stop_event = asyncio.Event()

    def _startup(self):
        self.loop = asyncio.get_running_loop()
        self._process_queue_task = self.loop.create_task(self.process_queue())
        # self.loop.create_task(self.queue_size_reporter(10))
        self._ready_event.set()

    async def handle(self, message: Union[commands.Command, events.Event]):
        if self._stop_event.is_set():
            raise Exception("MessageBus is stopped")
        await self.queue.put(message)

    async def process_queue(self):
        while not self._stop_event.is_set():
            try:
                message = await asyncio.wait_for(self.queue.get(), timeout=1.0)
            except asyncio.TimeoutError:
                continue
            if isinstance(message, events.Event):
                await self.handle_event(message)
            elif isinstance(message, commands.Command):
                await self.handle_command(message)
            else:
                raise Exception(f"{message} was not an Event or Command")
            self.queue.task_done()
        logger.debug("MessageBus stopped")

    async def handle_event(self, event: events.Event):
        task = self.loop.create_task(self.handler_registry.handle_event(event, self, self.event_broker))
        self.submissions.add(task)
        task.add_done_callback(lambda _: self.submissions.remove(task))

    async def handle_command(self, command: commands.Command):
        task = self.loop.create_task(self.handler_registry.handle_command(command, self))
        self.submissions.add(task)
        task.add_done_callback(lambda _: self.submissions.remove(task))

    async def __aenter__(self):
        self._startup()
        return self

    async def __aexit__(self, exc_type=None, exc_val=None, exc_tb=None):
        await self.shutdown()

    async def shutdown(self):
        await self.queue.join()
        await asyncio.gather(*self.submissions)
        self._stop_event.set()
        await self._process_queue_task

    async def wait_until_ready(self):
        await self._ready_event.wait()

    async def queue_size_reporter(self, interval: float = 1.0):
        while not self._stop_event.is_set():
            logger.info(f"Queue size: {self.queue.qsize()}")
            await asyncio.sleep(interval)

    async def _entrypoint(self):
        self.loop = asyncio.get_event_loop()
        async with self:
            await self.wait_until_ready()
            await self._stop_event.wait()

    async def enqueue(self, message: Union[commands.Command, events.Event]):
        await self.queue.put(message)

    def enqueue_sync(self, message: Union[commands.Command, events.Event]):
        try:
            loop = asyncio.get_event_loop()
        except RuntimeError:
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
        loop.run_until_complete(self.handle(message))

    def start(self):
        asyncio.run(self._entrypoint())

    def override_dependency(self, dependency, value):
        self.handler_registry.override_dependency(dependency, value)



#   Copyright 2023 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import time
from typing import Mapping
from uuid import UUID

from eventyst._exceptions import SchemaRegistryException
from eventyst.schema import SchemaFormat, SchemaVersion
from eventyst.serialization.registry import SchemaCompatibilityMode, SchemaRegistryClient
from eventyst.utilities.log import get_child_logger

logger = get_child_logger(__name__)

SCHEMA_VERSION_NOT_FOUND_MSG = 'Schema version is not found.'
SCHEMA_NOT_FOUND_MSG = 'Schema is not found.'

DEFAULT_COMPATIBILITY_MODE: SchemaCompatibilityMode = 'BACKWARD'


def schema_name_from_arn(arn: str) -> str:
    return arn.split('/')[-1]


class GlueSchemaRegistryClient(SchemaRegistryClient):
    """Schema registry for AWS Glue Schema Registry."""

    def __init__(
        self,
        glue_client,
        registry_name: str = 'default-registry',
        max_wait_attempts: int = 10,
        wait_interval_seconds: float = 3,
    ):
        self.glue_client = glue_client
        self.registry_name = registry_name
        self.max_wait_attempts = max_wait_attempts
        self.wait_interval_seconds = wait_interval_seconds

    def get_schema_version(self, version_id: UUID) -> SchemaVersion:
        """Get a schema version from the registry by id.

        Arguments:
            version_id: the schema version's unique id.

        Returns:
            SchemaVersion
        """
        try:
            res = self.glue_client.get_schema_version(SchemaVersionId=str(version_id))
        except Exception as e:
            raise SchemaRegistryException(f'Failed to get schema version by id {version_id}') from e
        if res['SchemaVersionId'] is None or res['Status'] != 'AVAILABLE':
            raise SchemaRegistryException(f"Schema Found but status is {res['Status']}")
        return SchemaVersion(
            schema_name=schema_name_from_arn(res['SchemaArn']),
            version_id=UUID(res['SchemaVersionId']),
            definition=res['SchemaDefinition'],
            data_format=res['DataFormat'],
            status=res['Status'],
            version_number=res['VersionNumber'],
        )

    def get_schema_by_definition(self, definition: str, schema_name: str) -> SchemaVersion:
        """Get a schema version from the registry by schema definition.

        Arguments:
            definition: the stringified schema definition.
            schema_name: the name of the schema.

        Returns:
            SchemaVersion
        """
        try:
            logger.debug('Getting schema version id for: name = %s, definition = %s', schema_name, definition)
            res = self.glue_client.get_schema_by_definition(
                SchemaId={'SchemaName': schema_name, 'RegistryName': self.registry_name}, SchemaDefinition=definition
            )
            if res['SchemaVersionId'] is None or res['Status'] != 'AVAILABLE':
                raise SchemaRegistryException(f"Schema Found but status is {res['Status']}")
            return SchemaVersion(
                schema_name=schema_name_from_arn(res['SchemaArn']),
                version_id=UUID(res['SchemaVersionId']),
                definition=definition,
                data_format=res['DataFormat'],
                status=res['Status'],
            )
        except Exception as e:
            raise SchemaRegistryException(
                'Failed to get schemaVersionId by schema definition for schema' f' name = {schema_name}'
            ) from e

    def get_or_register_schema_version(
        self,
        definition: str,
        schema_name: str,
        data_format: SchemaFormat,
        compatibility_mode: SchemaCompatibilityMode = DEFAULT_COMPATIBILITY_MODE,
        metadata: Mapping[str, str] = None,
    ) -> SchemaVersion:
        """Get Schema Version ID by following below steps:

        1) If schema version id exists in registry then get it from registry
        2) If schema version id does not exist in registry
             then if schema exists but version doesn't exist
                 then
                 2.1) Register schema version
             else if schema does not exist
                 then
                 2.2) create schema and register schema version

        Arguments:
            definition: the stringified schema definition.
            schema_name: the name of the schema in the registry.
            data_format: which format to use if creating the schema.
                Has no effect if the schema by name already exists.
            compatibility_mode: which compatibility mode to use if
                creating the schema. Has no effect if the schema by
                name already exists.
            metadata: optional metadata to add to the schema version
                if registering a new version. Has no effect if a
                schema version matching the specified definition already
                exists.
        """
        try:
            version = self.get_schema_by_definition(definition, schema_name)
        except SchemaRegistryException as e:
            cause_msg = str(e.__cause__)
            if SCHEMA_VERSION_NOT_FOUND_MSG in cause_msg:
                logger.debug(cause_msg)
                version_id = self.register_schema_version(definition, schema_name, metadata)
            elif SCHEMA_NOT_FOUND_MSG in cause_msg:
                logger.debug(cause_msg)
                version_id = self.create_schema(schema_name, data_format, definition, compatibility_mode, metadata)
            else:
                raise SchemaRegistryException(
                    'Exception occurred while fetching or registering schema'
                    f' definition = {definition}, schema name = {schema_name}'
                ) from e
            version = self.get_schema_version(version_id)
        return version

    def register_schema_version(self, definition: str, schema_name: str, metadata: Mapping[str, str] = None) -> UUID:
        """Register a new version to an existing schema.

        Waits until the new version becomes available before returning.

        Arguments:
            definition: the schema definition.
            schema_name: the name of the schema.
            metadata (optional): version metadata key-value pairs.

        Returns:
            UUID: the id of the new schema version
        """
        try:
            res = self.glue_client.register_schema_version(
                SchemaId={'SchemaName': schema_name, 'RegistryName': self.registry_name}, SchemaDefinition=definition
            )
            version_id = UUID(res['SchemaVersionId'])
            logger.info(
                'Registered the schema version with schema version '
                'id = %s and with version number = %s and status %s',
                version_id,
                res['VersionNumber'],
                res['Status'],
            )
            if res['Status'] != 'AVAILABLE':
                self._wait_for_schema_evolution_check_to_complete(version_id)
        except Exception as e:
            raise SchemaRegistryException(
                'Register schema :: Call failed when registering the schema'
                f' with the schema registry for schema name = {schema_name}',
            ) from e
        if metadata:
            self.put_schema_version_metadata(version_id, metadata)
        return version_id

    def _wait_for_schema_evolution_check_to_complete(self, schema_version_id: UUID):
        time.sleep(self.wait_interval_seconds)
        for _ in range(self.max_wait_attempts):
            res = self.glue_client.get_schema_version(SchemaVersionId=str(schema_version_id))
            status = res['Status']
            if status == 'AVAILABLE':
                break
            elif status != 'PENDING':
                raise SchemaRegistryException(
                    'Schema evolution check failed.' f' schemaVersionId {schema_version_id} is in' f' {status} status.'
                )
        else:
            raise SchemaRegistryException(
                'Retries exhausted for schema evolution check for ' f'schemaVersionId = {schema_version_id}'
            )

    def put_schema_version_metadata(self, version_id: UUID, metadata: Mapping[str, str]):
        for k, v in metadata.items():
            try:
                self.glue_client.put_schema_version_metadata(
                    SchemaVersionId=str(version_id), MetadataKeyValue={'MetadataKey': k, 'MetadataValue': v}
                )
            except Exception as e:
                raise SchemaRegistryException(
                    'Put schema version metadata :: Call failed when put'
                    f' metadata key = {k} value = {v} to schema for schema'
                    f' versionid = {version_id}'
                ) from e

    def create_schema(
        self,
        name: str,
        schema_format: SchemaFormat,
        definition: str,
        compatibility_mode: SchemaCompatibilityMode = DEFAULT_COMPATIBILITY_MODE,
        metadata: Mapping[str, str] = None,
    ) -> UUID:
        """Create a new schema and return the version id."""
        try:
            logger.info('Creating schema with name: %s and definition: %s', name, definition)
            res = self.glue_client.create_schema(
                SchemaName=name,
                RegistryId={'RegistryName': self.registry_name},
                DataFormat=schema_format,
                Compatibility=compatibility_mode,
                Description='',
                Tags={},
                SchemaDefinition=definition,
            )
            version_id = UUID(res['SchemaVersionId'])
            if metadata:
                self.put_schema_version_metadata(version_id, metadata)
        except Exception as e:
            if type(e).__name__ == 'AlreadyExistsException':
                logger.warning(
                    'Schema is already created, this could be caused by '
                    'multiple producers racing to auto-create schema.'
                )
                version_id = self.register_schema_version(definition, name, metadata)
            else:
                raise SchemaRegistryException(f'Create schema {name} failed') from e
        return version_id



#   Copyright 2023 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Welcome to Eventyst!"""
import hashlib
from uuid import UUID

from eventyst.core.application import Eventyst  # noqa: F401
from eventyst.core.commands import Command  # noqa: F401
from eventyst.core.dependency import dependency  # noqa: F401
from eventyst.core.events import Event  # noqa: F401
from eventyst.core.message_bus import MessageBus  # noqa: F401

__author__ = "Michael Statt"
__email__ = "michael.statt@modelyst.io"
__maintainer__ = "Michael Statt"
__maintainer_email__ = "michael.statt@modelyst.io"
__version__ = "0.1.2"


version_uuid = UUID(hashlib.md5(__version__.encode()).hexdigest())



#   Copyright 2023 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import os
import sys

import typer
from rich.traceback import Traceback

from eventyst.cli.styles import bad_typer_print, typer_print
from eventyst.core.application import Eventyst

# Errors
ERROR_FORMAT = "Application is not in MODULE:PACKAGE format: {0}"
ERROR_DEFAULT_APP = (
    "Could not find app at app.main:make_app.\nThis is the default --app value, did you forget to set --app?"
)
ERROR_MODULE = "Could not find module:\nModule: {0}\nPackage: {1}\nError: {2}"
ERROR_PACKAGE = "Could not find package within module:\nModule: {0}\nPackage: {1}\nError: {2}"
ERROR_ATTR = "Could not find Attribute within package:\nModule: {0}\nPackage: {1}\nError: {2}"
ERROR_NOT_APP = "Import String is not for a DBgen Model: \nImport String: {0}\nClass: {1}"
ERROR_NOT_APP_FUNCTION = (
    "Import String is for a function that does not produce a DBgen Model: \nImport String: {0}\nOutput Class: {1}"
)
ERROR_RUNNING_APP_FACT = (
    "Import String is for a function produced an error or required arguments: \nImport String: {0}\n"
    "Output Class: {1} \n{2}{3}"
)


def validate_app_str(app_str: str) -> Eventyst:
    """
    Validate the user input app import str checking for malformed and invalid inputs.py

    Args:
        app_str (str): CLI/config file input in MODULE:PACKAGE format

    Raises:
        typer.BadParameter: Whenever one of many checks is failed

    Returns:
        Eventyst: the output app after checks are in place
    """
    # Add current workind directory to the path at the end
    cwd = os.getcwd()
    sys.path.insert(0, cwd)

    def basic_error(fmt, val):
        return typer.BadParameter(fmt.format(*val), param_hint="--app")

    if app_str is None:
        raise typer.BadParameter("--app is required.")

    # Check for empty string app_strs
    if app_str in (":", ""):
        raise basic_error(ERROR_FORMAT, [app_str])

    split_app = app_str.split(":")
    if len(split_app) != 2:
        raise basic_error(ERROR_FORMAT, [app_str])
    module, package = split_app
    try:
        _temp = __import__(module, globals(), locals(), [package])
        app = getattr(_temp, package)
        if isinstance(app, Eventyst):
            return app
        elif callable(app):
            try:
                app = app()
            except TypeError as exc:
                print(exc)
                import traceback

                exc_str = traceback.format_exc()
                raise basic_error(
                    ERROR_RUNNING_APP_FACT, [app_str, type(app).__name__, "#" * 24 + "\n", str(exc_str)]
                ) from exc
            if isinstance(app, Eventyst):
                return app
            raise basic_error(ERROR_NOT_APP_FUNCTION, [app_str, type(app).__name__])

        raise basic_error(ERROR_NOT_APP, [app_str, type(app).__name__])
    except (ModuleNotFoundError, AttributeError) as exc:
        if app_str == 'app.main:make_app':
            raise basic_error(ERROR_DEFAULT_APP, []) from exc
        if f"No module named {module!r}" in str(exc):
            raise basic_error(ERROR_MODULE, [module, package, str(exc)]) from exc
        if isinstance(exc, AttributeError):
            typer_print('red')(Traceback(width=150))
            raise basic_error(ERROR_ATTR, [module, package, str(exc)]) from exc
        bad_typer_print(f"Error loading app at location '{module}.{package}'...")
        typer_print('red')(Traceback(width=150))
        raise typer.Exit(code=1)
    except Exception:
        bad_typer_print(f"Error loading app at location '{module}.{package}'...")
        typer_print('red')(Traceback(width=150))
        raise typer.Exit(code=1)



#   Copyright 2023 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import networkx as nx
from pyvis.network import Network

from eventyst.core.handler_registry import HandlerRegistry


def generate_graph(handler_registry: HandlerRegistry):
    G = nx.DiGraph()  # Create a directed graph

    # Add nodes and edges for command handlers
    for command, tracker in handler_registry.command_handlers.items():
        command_str = str(command.__name__)
        tracker_str = str(tracker.call.__name__)

        command_color = 'green' if command in handler_registry.entrypoints else 'lightblue'
        G.add_node(command_str, shape='square', color=command_color)

        # Add the command node and set its attributes
        G.add_node(tracker_str, shape='diamond', color='red')

        # Add an edge from the handler to the command
        G.add_edge(command_str, tracker_str)
        for event in tracker.emitted_types:
            event_str = str(event.__name__)
            # Add the event node and set its attributes
            G.add_node(event_str, shape='circle', color='orange')
            # Add an edge from the command to the emitted event
            G.add_edge(
                tracker_str,
                event_str,
            )

            # Mark broker events
            if event in handler_registry.broker_events:
                G.nodes[event_str]['color'] = 'orange'
                G.add_edge(event_str, 'Broker', color='orange')

    # Add nodes and edges for event handlers
    for event, trackers in handler_registry.event_handlers.items():
        event_str = str(event.__name__)
        # Add the event node and set its attributes
        event_color = 'orange' if event in handler_registry.broker_events else 'lightblue'

        if event in handler_registry.broker_events:
            G.add_edge(event_str, 'Broker', color='orange')

        if event in handler_registry.entrypoints:
            event_color = 'green'
        else:
            event_color = 'lightblue'

        G.add_node(event_str, shape='circle', color=event_color)

        # Add an edge from the event to the handler
        for tracker in trackers:
            tracker_str = str(tracker.call.__name__)
            G.add_node(tracker_str, shape='diamond', color='red')
            G.add_edge(event_str, tracker_str, color='lightblue')

            for emitted_event in tracker.emitted_types:
                emitted_event_str = str(emitted_event.__name__)
                # Add the emitted event node and set its attributes
                G.add_node(
                    emitted_event_str,
                    shape='circle',
                    color='lightblue',
                )

                # Add an edge from the handler to the emitted event
                G.add_edge(
                    tracker_str,
                    emitted_event_str,
                )
                if emitted_event in handler_registry.broker_events:
                    G.add_edge(emitted_event_str, 'Broker', color='orange')
    if handler_registry.broker_events:
        G.add_node('Broker', shape='oval', color='orange', size=20)

    return G


def draw_graph_pyvis(G):
    net = Network(directed=True, notebook=True, cdn_resources='in_line', layout='hierarchical')
    net.from_nx(G)
    return net



#   Copyright 2023 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Utilities for printing things to screen for CLI."""
import typer
from rich.console import Console, Group
from rich.panel import Panel
from rich.text import Text
from rich.theme import Theme

from eventyst import __version__

THEME_COLOR = typer.colors.GREEN
theme = Theme({'theme': 'magenta'})
console = Console(theme=theme)
LOGO = """
███████╗██╗   ██╗███████╗███╗   ██╗████████╗██╗   ██╗███████╗████████╗
██╔════╝██║   ██║██╔════╝████╗  ██║╚══██╔══╝╚██╗ ██╔╝██╔════╝╚══██╔══╝
█████╗  ██║   ██║█████╗  ██╔██╗ ██║   ██║    ╚████╔╝ ███████╗   ██║
██╔══╝  ╚██╗ ██╔╝██╔══╝  ██║╚██╗██║   ██║     ╚██╔╝  ╚════██║   ██║
███████╗ ╚████╔╝ ███████╗██║ ╚████║   ██║      ██║   ███████║   ██║
╚══════╝  ╚═══╝  ╚══════╝╚═╝  ╚═══╝   ╚═╝      ╚═╝   ╚══════╝   ╚═╝"""

VERSION = f"""VERSION: {__version__}"""


def delimiter(color: str = THEME_COLOR):
    console.rule(style=color)


LOGO_STYLE = Panel.fit(
    Group(Panel(Text(LOGO)), Panel(Text(VERSION, justify='center'))),
    style='theme',
)


# Easy printers
def typer_print(color=None):
    return lambda msg: console.print(msg, style=color)


good_typer_print = typer_print('green')
bad_typer_print = typer_print('red')
theme_typer_print = typer_print('theme')



#   Copyright 2023 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import functools
from typing import Any, Callable, NamedTuple
from uuid import UUID

from eventyst.schema import AvroSchema, BaseSchema
from eventyst.serialization.codec import decode, encode
from eventyst.serialization.naming import SchemaNamingStrategy, topic_name_strategy
from eventyst.serialization.registry import SchemaCompatibilityMode, SchemaRegistryClient, SchemaVersion
from eventyst.utilities.log import get_child_logger

logger = get_child_logger(__name__)


class DataAndSchema(NamedTuple):
    """Data and its schema.

    Can be used to wrap the data and schema together before calling the
    producer's producing methods.
    """

    data: Any
    schema: BaseSchema


class Serializer:
    """Kafka serializer that uses the AWS Schema Registry.

    Arguments:
        client: instance of SchemaRegistryClient
        is_key (optional): whether the serializer is serializing keys as
            opposed to values. Defaults to false. Setting this to the
            appropriate value is important to avoid mixing key and value
            schemas if using the default schema name strategy.
        compatibility_mode (optional): the compatibility mode t use if
            creating a new schema in the registry. Defaults to the
            registry's default compatibility setting if not specified.
        schema_naming_strategy (optional): how to choose the schema name
            when creating new schemas. Defaults to the topic name
            strategy. See the `naming` module for more information and
            alternate strategies.
    """

    def __init__(
        self,
        client: SchemaRegistryClient,
        is_key: bool = False,
        compatibility_mode: SchemaCompatibilityMode = 'BACKWARD',
        schema_naming_strategy: SchemaNamingStrategy = topic_name_strategy,
    ):
        self.client = client
        self.is_key = is_key
        self.compatibility_mode: SchemaCompatibilityMode = compatibility_mode
        self.schema_naming_strategy = schema_naming_strategy

    def get_serializer(self, topic: str) -> Callable[[DataAndSchema], bytes]:
        """Returns a serializer function for the given topic.

        The serializer function takes a single argument, the data to
        serialize, and returns the serialized bytes.

        Arguments:
            topic: the topic to serialize data for
        """
        return functools.partial(self.serialize, topic)

    def serialize(self, topic: str, data_and_schema: DataAndSchema) -> bytes:
        if data_and_schema is None:
            return None
        if not isinstance(data_and_schema, tuple):
            raise TypeError('KafkaSerializer can only serialize', f' {tuple}, got {type(data_and_schema)}')
        data, schema = data_and_schema
        schema_version = self._get_schema_version(topic, schema)
        serialized = schema.write(data)
        return encode(serialized, schema_version.version_id)

    @functools.lru_cache(maxsize=None)
    def _get_schema_version(self, topic: str, schema: BaseSchema) -> SchemaVersion:
        schema_name = self.schema_naming_strategy(topic, self.is_key, schema)
        logger.info('Fetching schema %s...', schema_name)
        return self.client.get_or_register_schema_version(
            definition=schema.json(),
            schema_name=schema_name,
            data_format=schema.data_format,
            compatibility_mode=self.compatibility_mode,
        )


class Deserializer:
    """Kafka serializer that uses a Schema Registry.

    Arguments:
        client: instance of SchemaRegistryClient.
        secondary_deserializer: optional deserializer to pass through
            to when processing values with an unrecognized encoding.
            This is primarily use to migrate from other schema
            registries or handle schema-less data. The secondary deserializer
            should either be a callable taking the same arguments as
            deserialize or an object with a matching deserialize method.
    """

    def __init__(
        self,
        client: SchemaRegistryClient,
        reader_schema: BaseSchema = None,
        reader_schema_id: UUID = None,
    ):
        self.client = client
        self.reader_schema = reader_schema
        self.schema_version_id = reader_schema_id

    def __call__(self, bytes_: bytes) -> DataAndSchema:
        return self.deserialize(bytes_)

    def deserialize(self, bytes_: bytes) -> DataAndSchema:
        if bytes_ is None:
            return None
        data_bytes, schema_version_id = decode(bytes_)
        writer_schema_version = self._get_schema_version(schema_version_id)
        writer_schema = self._schema_for_version(writer_schema_version)

        # If the reader schema is not set, use the writer schema.
        if self.reader_schema:
            reader_schema = self.reader_schema
        elif self.schema_version_id:
            reader_schema_version = self._get_schema_version(schema_version_id)
            reader_schema = self._schema_for_version(reader_schema_version)
        else:
            reader_schema = None

        # If the reader schema is set, validate the writer schema against it.
        message = writer_schema.read(data_bytes, reader_schema=reader_schema)

        return DataAndSchema(message, reader_schema)

    @functools.lru_cache(maxsize=None)
    def _get_schema_version(self, version_id: UUID):
        return self.client.get_schema_version(version_id)

    @functools.lru_cache(maxsize=None)
    def _schema_for_version(self, version: SchemaVersion) -> BaseSchema:
        if version.data_format == 'AVRO':
            return AvroSchema(version.definition)
        raise NotImplementedError("Only AVRO is supported at the moment.")



#   Copyright 2023 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import datetime
import io
import json
from typing import Any
from uuid import UUID, uuid4, uuid5

import avro.errors
import avro.schema
from avro.io import BinaryDecoder, BinaryEncoder, DatumReader, DatumWriter
from pydantic import BaseModel, create_model

from eventyst.configuration import settings
from eventyst.schema import AvroSchema
from eventyst.utilities.avro_schema import schema


class EventMetadata(BaseModel):
    id: UUID
    created_at: float
    event_type: str
    source: str
    source_ids: list[UUID] | None = None


class BaseMessage(BaseModel):
    __namespace__ = "eventyst.core.message"

    @classmethod
    def avro_schema(cls) -> AvroSchema:
        envelope = create_model(
            cls.__name__ + "Envelope",
            metadata=(EventMetadata, ...),
            payload=(cls, ...),
        )
        # TODO solve issue with list arrays
        return AvroSchema(schema(envelope, namespace=cls.__namespace__))

    @classmethod
    def topic(cls) -> str:
        suffix = getattr(cls, "__topic__", cls.__name__)
        return settings.SERVICE_NAME + f"-{suffix}"

    def envelope(self, source_ids: list[UUID] | None = None) -> dict[str, Any]:
        now = datetime.datetime.now()
        return {
            "metadata": EventMetadata(
                id=uuid5(uuid4(), settings.SERVICE_NAME),
                created_at=now.timestamp(),
                event_type=self.__class__.__name__,
                source=settings.SERVICE_NAME,
                source_ids=source_ids,
            ).dict(),
            "payload": self.dict(),
        }

    def serialize(self) -> bytes:
        # Serialize the object to Avro format
        schema = avro.schema.parse(json.dumps(self.avro_schema()))
        writer = DatumWriter(schema)
        bytes_writer = io.BytesIO()
        encoder = BinaryEncoder(bytes_writer)
        writer.write(self.envelope(), encoder)
        return bytes_writer.getvalue()

    @classmethod
    def deserialize(cls, serialized_data):
        # Deserialize Avro data to create a new object of the subclass
        schema = avro.schema.parse(json.dumps(cls.avro_schema()))
        reader = DatumReader(schema)
        bytes_reader = io.BytesIO(serialized_data)
        decoder = BinaryDecoder(bytes_reader)
        try:
            data = reader.read(decoder)
        except avro.errors.InvalidAvroBinaryEncoding as e:
            raise ValueError(f"Invalid Avro encoding: {e}")

        if not isinstance(data, dict):
            raise ValueError(f"Invalid Avro encoding: {data}")

        obj = cls.parse_obj(data["payload"])
        metadata = EventMetadata.parse_obj(data["metadata"])
        return obj, metadata


class Event(BaseMessage):
    """Base class for all events."""

    __namespace__ = "eventyst.core.events"



#   Copyright 2023 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import inspect
import sys
from typing import Any, Callable, Tuple, Type, Union

try:
    from typing import GenericAlias as TypingGenericAlias  # type: ignore
except ImportError:
    # python < 3.9 does not have GenericAlias (list[int], tuple[str, ...] and so on)
    TypingGenericAlias = ()  # type: ignore

if sys.version_info < (3, 10):
    WithArgsTypes = (TypingGenericAlias,)
else:
    import types

    WithArgsTypes = (types.GenericAlias, types.UnionType)


# Adapted from pydantic
def lenient_issubclass(cls: Any, class_or_tuple: Union[Type[Any], Tuple[Type[Any], ...], None]) -> bool:
    try:
        return isinstance(cls, type) and issubclass(cls, class_or_tuple)  # type: ignore[arg-type]
    except TypeError:
        if isinstance(cls, WithArgsTypes):
            return False
        raise  # pragma: no cover


def is_coroutine_callable(call: Callable[..., Any]) -> bool:
    if inspect.isroutine(call):
        return inspect.iscoroutinefunction(call)
    if inspect.isclass(call):
        return False
    dunder_call = getattr(call, "__call__", None)  # noqa: B004
    return inspect.iscoroutinefunction(dunder_call)



#   Copyright 2023 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from dataclasses import dataclass
from typing import Literal, Mapping, Optional, Protocol
from uuid import UUID

SchemaVersionStatus = Literal['AVAILABLE', 'DELETING', 'FAILURE', 'PENDING']
SchemaFormat = Literal['AVRO', 'JSON']
SchemaCompatibilityMode = Literal['BACKWARD', 'FORWARD', 'FULL', 'NONE']


@dataclass
class SchemaVersion:
    schema_name: str
    version_id: UUID
    definition: str
    data_format: SchemaFormat
    status: SchemaVersionStatus
    version_number: Optional[int] = None

    def __hash__(self):
        return hash(self.definition)


class SchemaRegistryClient(Protocol):
    """Protocol for schema registries."""

    def get_schema_version(self, version_id: UUID) -> SchemaVersion:
        ...

    def register_schema_version(self, schema_name: str, definition: str) -> SchemaVersion:
        ...

    def create_schema(self, schema_name: str, definition: str) -> SchemaVersion:
        ...

    def get_schema_by_definition(self, schema_name: str, definition: str) -> SchemaVersion:
        ...

    def get_or_register_schema_version(
        self,
        schema_name: str,
        definition: str,
        data_format: SchemaFormat,
        compatibility_mode: SchemaCompatibilityMode,
    ) -> SchemaVersion:
        ...

    def put_schema_version_metadata(version_id: UUID, metadata: Mapping[str, str]) -> None:
        ...
